# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

import numpy
import pandas
import tensorflow as tf
from mlgb.error import MLGBError


def check_filepath(*filepath):
    for fp in filepath:
        if not os.path.exists(fp):
            os.mkdir(fp)
    return


def get_dense_feature_name_dict(feature_name):
    return {'feature_name': feature_name}


def get_sparse_feature_name_dict(feature_name, feature_nunique, input_length, embed_dim=None, embed_alpha=6, mask_zero=True):
    # embed_dim: https://developers.googleblog.com/2017/11/introducing-tensorflow-feature-columns.html
    # embed_alpha=6: when embed_feature_nunique is in MinMax[4, 1e4], then embed_dim in MinMax[8, 60].
    name_dict = {
        'feature_name': feature_name,
        'input_length': input_length,
        'mask_zero': mask_zero,
        'feature_nunique': feature_nunique,
        'embed_feature_nunique': feature_nunique + 1 if mask_zero else feature_nunique,
        'embed_dim': embed_dim if embed_dim else int((feature_nunique + 1) ** 0.25 * embed_alpha),
    }
    return name_dict


def get_embed_width(sparse_feature_names, embed_dim):
    if embed_dim:
        return int(len(sparse_feature_names) * embed_dim)
    return int(sum([feat_dict['embed_dim'] for feat_dict in sparse_feature_names]))


def get_onehot_width(sparse_feature_names):
    return sum([feat_dict['feature_nunique'] for feat_dict in sparse_feature_names])


def get_padded_sequence(sequential_inputs, mode='numpy', sequence_max_length=8, sequence_pad_value=0.0):
    if mode not in ('numpy', 'pandas'):
        raise MLGBError

    _, n_sequences = sequential_inputs.shape

    sequence_pool = []
    for i in range(n_sequences):
        sequence_i = tf.keras.utils.pad_sequences(
            sequences=sequential_inputs[:, i] if mode == 'numpy' else sequential_inputs.iloc[:, i],
            maxlen=sequence_max_length,
            padding='pre',  # Union['pre', 'post']
            truncating='pre',
            value=sequence_pad_value,
        )
        sequence_pool.append(sequence_i)

    if mode == 'numpy':
        sequential_outputs = numpy.stack(sequence_pool, axis=1)  # (batch_size, features_width, sequence_length)
    else:
        sequential_outputs = pandas.DataFrame({
            sequential_inputs.columns[i]: sequence_pool[i].tolist() for i in range(n_sequences)
        })
    return sequential_outputs





# coding=utf-8
# author=uliontse

import numpy
import torch
from sklearn.metrics import roc_auc_score, f1_score, log_loss
from sklearn.metrics import confusion_matrix, classification_report

from mlgb import get_model, ranking_models
from mlgb.data import get_binary_label_data
from mlgb.utils import check_filepath


def train_and_evaluate(model, optimizer, loss_fn, x_train, y_train, x_test, y_test,
                       epochs=10, train_batch_size=32, test_batch_size=32, valid_ratio=0.15):

    train_batch_num = int(len(y_train) // train_batch_size)  # drop_last=True
    test_batch_num = int(len(y_test) // test_batch_size)  # drop_last=True

    print(end='\n')
    for epoch in range(epochs):
        model.train()
        for i in range(train_batch_num):
            id_l, id_r = int(train_batch_size * i), int(train_batch_size * (i + 1 - valid_ratio))
            if model_name in two_inputs_models:
                x_train_batch = [[m[id_l: id_r] for m in x_t] for x_t in x_train]
            else:
                x_train_batch = [m[id_l: id_r] for m in x_train]
            y_train_batch = y_train[id_l: id_r]

            y_pred_batch = torch.squeeze(model(x_train_batch))
            y_train_batch = torch.as_tensor(y_train_batch, dtype=torch.float32, device=device)

            l1l2_loss = model.l1l2_loss()
            loss = loss_fn(y_pred_batch, y_train_batch) + l1l2_loss  #
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

        model.eval()
        y_train_epoch = numpy.array([])
        y_valid_epoch = numpy.array([])
        y_test_epoch = numpy.array([])
        y_train_pred_prob_epoch = numpy.array([])
        y_valid_pred_prob_epoch = numpy.array([])
        y_test_pred_prob_epoch = numpy.array([])
        with torch.no_grad():
            for i in range(train_batch_num):
                id_l, id_r = int(train_batch_size * i), int(train_batch_size * (i + 1 - valid_ratio))
                if model_name in two_inputs_models:
                    x_train_batch = [[m[id_l: id_r] for m in x_t] for x_t in x_train]
                else:
                    x_train_batch = [m[id_l: id_r] for m in x_train]
                y_train_batch = y_train[id_l: id_r]

                y_train_output = torch.squeeze(model(x_train_batch)).cpu().numpy()
                y_train_pred_prob_epoch = numpy.concatenate([y_train_pred_prob_epoch, y_train_output], axis=0)
                y_train_epoch = numpy.concatenate([y_train_epoch, y_train_batch], axis=0)

            for i in range(train_batch_num):
                id_l, id_r = int(train_batch_size * (i + 1 - valid_ratio)), int(train_batch_size * (i + 1))
                if model_name in two_inputs_models:
                    x_valid_batch = [[m[id_l: id_r] for m in x_t] for x_t in x_train]
                else:
                    x_valid_batch = [m[id_l: id_r] for m in x_train]
                y_valid_batch = y_train[id_l: id_r]

                y_valid_output = torch.squeeze(model(x_valid_batch)).cpu().numpy()
                y_valid_pred_prob_epoch = numpy.concatenate([y_valid_pred_prob_epoch, y_valid_output], axis=0)
                y_valid_epoch = numpy.concatenate([y_valid_epoch, y_valid_batch], axis=0)

            for i in range(test_batch_num):
                id_l, id_r = int(train_batch_size * i), int(train_batch_size * (i + 1))
                if model_name in two_inputs_models:
                    x_test_batch = [[m[id_l: id_r] for m in x_t] for x_t in x_test]
                else:
                    x_test_batch = [m[id_l: id_r] for m in x_test]
                y_test_batch = y_test[id_l: id_r]

                y_test_output = torch.squeeze(model(x_test_batch)).cpu().numpy()
                y_test_pred_prob_epoch = numpy.concatenate([y_test_pred_prob_epoch, y_test_output], axis=0)
                y_test_epoch = numpy.concatenate([y_test_epoch, y_test_batch], axis=0)

        y_train_pred_epoch = numpy.where(y_train_pred_prob_epoch >= 0.5, 1, 0)
        y_valid_pred_epoch = numpy.where(y_valid_pred_prob_epoch >= 0.5, 1, 0)
        y_test_pred_epoch = numpy.where(y_test_pred_prob_epoch >= 0.5, 1, 0)
        result = {
            'epoch': epoch + 1,
            'bce': round(log_loss(y_train_epoch, y_train_pred_prob_epoch), 4),
            'auc': round(roc_auc_score(y_train_epoch, y_train_pred_prob_epoch), 4),
            'f1': round(f1_score(y_train_epoch, y_train_pred_epoch, average='macro'), 4),

            'val_bce': round(log_loss(y_valid_epoch, y_valid_pred_prob_epoch), 4),
            'val_auc': round(roc_auc_score(y_valid_epoch, y_valid_pred_prob_epoch), 4),
            'val_f1': round(f1_score(y_valid_epoch, y_valid_pred_epoch, average='macro'), 4),

            'test_bce': round(log_loss(y_test_epoch, y_test_pred_prob_epoch), 4),
            'test_auc': round(roc_auc_score(y_test_epoch, y_test_pred_prob_epoch), 4),
            'test_f1': round(f1_score(y_test_epoch, y_test_pred_epoch, average='macro'), 4),
        }
        print('evaluate:', result)
    return model


if __name__ == '__main__':
    # model_name = 'PNN'
    lang = 'torch'
    device = 'cuda'
    seed = 0

    two_inputs_models = ['PLM', 'GRU4Rec', 'Caser', 'SASRec', 'BERT4Rec', 'BST', 'DIN', 'DIEN', 'DSIN']

    data = get_binary_label_data(
        n_samples=int(1e3),
        negative_class_weight=0.9,
        test_size=0.15,
        inputs_if_2_groups=False,
        seed=seed,
    )
    plm_data = get_binary_label_data(
        n_samples=int(1e3),
        negative_class_weight=0.9,
        test_size=0.15,
        inputs_if_2_groups=True,
        seed=seed,
    )

    for i, model_name in enumerate(ranking_models[41:]):
        print(i, model_name, end='\n\n')

        # path of save_model:
        tmp_dir = '.tmp'
        model_dir = f'{tmp_dir}/{model_name}_{lang}'
        log_dir = f'{model_dir}/log_dir'
        save_model_dir = f'{model_dir}/save_model'
        save_model_file = f'{save_model_dir}/model.pth'
        check_filepath(tmp_dir, model_dir, log_dir, save_model_dir,)

        # get_data:
        feature_names, (x_train, y_train), (x_test, y_test) = plm_data if model_name in two_inputs_models else data
        if model_name in two_inputs_models:
            print(f'features: {[[len(names) for names in group] for group in feature_names]}', end='\n\n')
        else:
            print(f'features: {[len(names) for names in feature_names]}', end='\n\n')

        # train and evaluate:
        model = get_model(
            feature_names=feature_names,
            model_name=model_name,
            task='binary',
            aim='ranking',
            lang='torch',
            device=device,
            seed=seed,
            model_l1=0.0,
            model_l2=1e-6,
        )
        print(model, end='\n\n')

        # must init parameters by build() like tf at first run:
        _ = model(x_test)  # shape requires: any batch_size, other shape must be same as x_train and x_test.
        for name, p in model.named_parameters():
            print(name, list(p.shape), sep='\t\t')

        optimizer = torch.optim.NAdam(model.parameters(), lr=1e-3, weight_decay=0.0)  # model_l2 vs weight_decay
        loss_fn = torch.nn.BCELoss()
        model = train_and_evaluate(model, optimizer, loss_fn,
                                   x_train, y_train, x_test, y_test,
                                   epochs=10, train_batch_size=32, test_batch_size=32, valid_ratio=0.15)

        y_pred_prob = numpy.squeeze(model(x_test).detach().cpu().numpy())
        y_pred = numpy.where(y_pred_prob >= 0.5, 1, 0)
        print(y_pred.shape)
        print(y_test.shape)

        print(confusion_matrix(y_test, y_pred))
        print(classification_report(y_test, y_pred))

        # ranking model:
        torch.save(obj=model, f=save_model_file)
        local_model = torch.load(f=save_model_file)

        y_local_pred = numpy.squeeze(local_model(x_test).detach().cpu().numpy())
        y_model_pred = numpy.squeeze(model(x_test).detach().cpu().numpy())
        print('y_local_pred == y_model_pred:', numpy.allclose(y_local_pred, y_model_pred))
        print(numpy.abs(y_local_pred - y_model_pred).max())

        print('done', end='\n\n')







# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    tf,
    L1L2,
    Dropout,
    LayerNormalization,
)
from mlgb.tf.functions import (
    ActivationLayer,
    InitializerLayer,
    MaskLayer,
    PositionalEncoding,
)
from mlgb.error import MLGBError


__all__ = [
    'SimpleAttentionLayer',
    'LabelAttentionLayer',
    'TransformerLayer',
    'InteractingLayer',
]


class SimpleAttentionLayer(tf.keras.layers.Layer):
    def __init__(self, softmax_axis=1, softmax_if_mask=True, softmax_pre_temperature_ratio=1.0):
        super().__init__()
        if not (1e-2 <= softmax_pre_temperature_ratio <= 1.0):
            raise MLGBError

        self.softmax_axis = softmax_axis
        self.softmax_pre_temperature_ratio = softmax_pre_temperature_ratio
        self.mask_fn = MaskLayer(att_if_mask=softmax_if_mask)

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank not in (2, 3):
            raise MLGBError
        if input_shape[0] != input_shape[1]:
            raise MLGBError

        self.scale = input_shape[0][-1] ** 0.5  #
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_q, x_k = inputs
        x_v = x_k

        w = x_q * x_k / self.scale
        w = w / self.softmax_pre_temperature_ratio
        w = self.mask_fn(w)
        w = tf.nn.softmax(w, axis=self.softmax_axis)
        x_v = x_v * w
        return x_v


class LabelAttentionLayer(SimpleAttentionLayer):
    def __init__(self, softmax_axis=1, softmax_pre_temperature_ratio=1.0):
        super().__init__(softmax_axis, True, softmax_pre_temperature_ratio)


class MultiHeadAttentionLayer(tf.keras.layers.Layer):
    def __init__(self, mha_head_num=8, mha_head_dim=32, mha_if_mask=True, mha_l2=0.0, mha_initializer=None, seed=None):
        super().__init__()
        if not (mha_head_num > 0 and mha_head_dim > 0):
            raise MLGBError

        self.mha_head_num = mha_head_num
        self.mha_head_dim = mha_head_dim
        self.mha_model_dim = mha_head_num * mha_head_dim
        self.mha_l2 = mha_l2
        self.mha_initializer_list = [
            InitializerLayer(
                initializer=mha_initializer,
                activation=None,
                seed=seed + i if isinstance(seed, int) else seed,
            ).get() for i, _ in enumerate('qkv')
        ]
        self.mask_fn = MaskLayer(att_if_mask=mha_if_mask)

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 3):
            raise MLGBError
        if input_shape[0] != input_shape[1]:
            raise MLGBError

        _, self.fields_width, embed_dim = input_shape[1]

        self.qkv_input_weight_list = [
            self.add_weight(
                name=f'qkv_input_weight_{w}',
                shape=[embed_dim, self.mha_model_dim],
                initializer=self.mha_initializer_list[i],
                regularizer=L1L2(0.0, self.mha_l2),
                trainable=True,
            ) for i, w in enumerate('qkv')
        ]
        self.qkv_output_weight = self.add_weight(
            name=f'qkv_output_weight',
            shape=[self.mha_model_dim, embed_dim],
            initializer=self.mha_initializer_list[0],
            regularizer=L1L2(0.0, self.mha_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        # NLP inputs only has 1 vocabulary, 1 sequential feature, sequence_max_length == fields_width.
        # NLP's fields_width << CTR's fields_width, NLP's embed_dim >> CTR's embed_dim,
        # so there are different projection.

        x_q, x_k = inputs
        x_v = x_k  # (b, f, e)

        x_q @= self.qkv_input_weight_list[0]  # (batch_size, fields_width, model_dim)
        x_k @= self.qkv_input_weight_list[1]
        x_v @= self.qkv_input_weight_list[2]

        qkv_4d_shape = [-1, self.fields_width, self.mha_head_num, self.mha_head_dim]
        x_q = tf.transpose(tf.reshape(x_q, shape=qkv_4d_shape), perm=[0, 2, 1, 3])  # (b, head_num, f, head_dim)
        x_k = tf.transpose(tf.reshape(x_k, shape=qkv_4d_shape), perm=[0, 2, 1, 3])
        x_v = tf.transpose(tf.reshape(x_v, shape=qkv_4d_shape), perm=[0, 2, 1, 3])

        x_att = tf.einsum('bhid,bhjd->bhij', x_q, x_k) / (self.mha_head_dim ** 0.5)  # (b, head_num, f, f)
        x_att = self.mask_fn(x_att)
        x_att = tf.nn.softmax(x_att, axis=-1)  #
        x_v = tf.einsum('bhfi,bhfj->bhij', x_att, x_v)  # (b, head_num, f, head_dim)

        qkv_3d_shape = [-1, self.fields_width, self.mha_model_dim]
        x_v = tf.reshape(tf.transpose(x_v, perm=[0, 2, 1, 3]), shape=qkv_3d_shape)  # (b, f, model_dim)
        x_v @= self.qkv_output_weight  # (b, f, e)
        return x_v


class TransformerFeedForwardNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, ffn_activation='gelu', ffn_l2=0.0, ffn_initializer=None, seed=None):
        super().__init__()
        self.ffn_l2 = ffn_l2
        self.ffn_initializer_list = [
            InitializerLayer(
                initializer=ffn_initializer,
                activation=None,
                seed=seed + i if isinstance(seed, int) else seed,
            ).get() for i in range(2)
        ]
        self.activation_fn = ActivationLayer(activation=ffn_activation)

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        _, f, e = input_shape

        self.ffn_weight_list = [
            self.add_weight(
                name=f'ffn_weight_{i}',
                shape=[f, e],
                initializer=self.ffn_initializer_list[i],
                regularizer=L1L2(0.0, self.ffn_l2),
                trainable=True,
            ) for i in range(2)
        ]
        self.ffn_bias_list = [
            self.add_weight(
                name=f'ffn_bias_{i}',
                shape=[f, 1],
                initializer='zeros',
                regularizer=L1L2(0.0, self.ffn_l2),
                trainable=True,
            ) for i in range(2)
        ]
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = x * self.ffn_weight_list[0] + self.ffn_bias_list[0]
        x = self.activation_fn(x)
        x = x * self.ffn_weight_list[1] + self.ffn_bias_list[1]
        return x


class TransformerResidualLayer(tf.keras.layers.Layer):
    def __init__(self, residual_ln_axis=1, residual_dropout=0.0):
        super().__init__()
        self.ln_fn = LayerNormalization(axis=residual_ln_axis)
        self.dropout_fn = Dropout(rate=residual_dropout)

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 3):
            raise MLGBError
        if input_shape[0] != input_shape[1]:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x, x_stack = inputs

        x = tf.add(x, x_stack)
        x = self.ln_fn(x)
        x = self.dropout_fn(x)
        return x


class TransformerLayer(tf.keras.layers.Layer):
    def __init__(self, trm_if_pe=False, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_l2=0.0, trm_ffn_initializer=None,
                 trm_residual_dropout=0.0, seed=None):
        super().__init__()
        self.trm_if_ffn = trm_if_ffn

        self.pe_fn = PositionalEncoding(att_if_pe=trm_if_pe)
        self.mha_fn = MultiHeadAttentionLayer(
            mha_head_num=trm_mha_head_num,
            mha_head_dim=trm_mha_head_dim,
            mha_if_mask=trm_mha_if_mask,
            mha_l2=trm_mha_l2,
            mha_initializer=trm_mha_initializer,
            seed=seed,
        )
        if self.trm_if_ffn:
            self.ffn_fn = TransformerFeedForwardNetworkLayer(
                ffn_activation=trm_ffn_activation,
                ffn_l2=trm_ffn_l2,
                ffn_initializer=trm_ffn_initializer,
                seed=seed,
            )
        self.residual_fn_list = [
            TransformerResidualLayer(
                residual_ln_axis=1,
                residual_dropout=trm_residual_dropout,
            ) for _ in range(2 if self.trm_if_ffn else 1)
        ]

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 3):
            raise MLGBError
        if input_shape[0] != input_shape[1]:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_q, x_k = inputs

        x_q = self.pe_fn(x_q)
        x_k = self.pe_fn(x_k)
        x = self.mha_fn([x_q, x_k])
        x_k = self.residual_fn_list[0]([x, x_k])

        if self.trm_if_ffn:
            x = self.ffn_fn(x_k)
            x_k = self.residual_fn_list[1]([x, x_k])
        return x_k


class InteractingLayer(TransformerLayer):
    def __init__(self, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_residual_dropout=0.0, seed=None):
        super().__init__(
            trm_if_pe=False,
            trm_mha_head_num=trm_mha_head_num, trm_mha_head_dim=trm_mha_head_dim,
            trm_mha_if_mask=trm_mha_if_mask, trm_mha_l2=trm_mha_l2, trm_mha_initializer=trm_mha_initializer,
            trm_if_ffn=False,
            trm_residual_dropout=trm_residual_dropout, seed=seed,
        )



















# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    tf,
    numpy,
    L1L2,
    Flatten,
    SampleModeList
)
from mlgb.tf.functions import (
    InitializerLayer,
    ActivationLayer,
    MaskLayer,
)
from mlgb.tf.inputs import InputsLayer
from mlgb.tf.components.linears import DeepNeuralNetworkLayer
from mlgb.error import MLGBError


__all__ = [
    'SampledSoftmaxLossLayer',
    'BaseInputsEmbeddingLayer',
    'CapsuleNetworkLayer',
]


# @tf.keras.utils.register_keras_serializable(package='SampledSoftmaxLossLayer')
class SampledSoftmaxLossLayer(tf.keras.layers.Layer):
    def __init__(self, sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 seed=None, **kwargs):  # debug1: param kwargs is must. If not, load_model will show bug.
        super().__init__()
        if sample_mode not in SampleModeList:
            raise MLGBError

        self.sample_mode = sample_mode
        self.sample_num = sample_num
        self.sample_item_distribution_list = sample_item_distribution_list
        self.sample_fixed_unigram_frequency_list = sample_fixed_unigram_frequency_list
        self.sample_fixed_unigram_distortion = sample_fixed_unigram_distortion
        self.seed = seed

    @tf.function
    def call(self, y_true, y_pred):
        y_true = tf.convert_to_tensor(y_true, dtype=tf.float32)
        y_pred = tf.convert_to_tensor(y_pred, dtype=tf.float32)
        self.check_shape(y_true, y_pred)

        if self.sample_mode != 'Sample:all':
            y_true_int = tf.cast(y_true, dtype=tf.int64)
            _, multiclass_num = y_true_int.shape
            if self.sample_num and self.sample_num > multiclass_num:
                raise MLGBError('sample_num')
            self.sample_num = self.sample_num if self.sample_num else multiclass_num

            if self.sample_mode == 'Sample:uniform':
                item_indices = tf.random.uniform_candidate_sampler(
                    true_classes=y_true_int,
                    num_true=multiclass_num,
                    num_sampled=self.sample_num,
                    range_max=multiclass_num,
                    unique=True,
                    seed=self.seed,
                )[0]
            elif self.sample_mode == 'Sample:log_uniform':
                item_indices = tf.random.log_uniform_candidate_sampler(
                    true_classes=y_true_int,
                    num_true=multiclass_num,
                    num_sampled=self.sample_num,
                    range_max=multiclass_num,
                    unique=True,
                    seed=self.seed,
                )[0]
            elif self.sample_mode == 'Sample:fixed_unigram':
                if not self.sample_fixed_unigram_frequency_list:
                    self.sample_fixed_unigram_frequency_list = [1.0 / multiclass_num] * multiclass_num
                item_indices = tf.random.fixed_unigram_candidate_sampler(
                    true_classes=y_true_int,
                    num_true=multiclass_num,
                    num_sampled=self.sample_num,
                    range_max=multiclass_num,
                    unigrams=self.sample_fixed_unigram_frequency_list,
                    distortion=self.sample_fixed_unigram_distortion,
                    unique=True,
                    seed=self.seed,
                )[0]
            elif self.sample_mode == 'Sample:learned_unigram':
                item_indices = tf.random.learned_unigram_candidate_sampler(
                    true_classes=y_true_int,
                    num_true=multiclass_num,
                    num_sampled=self.sample_num,
                    range_max=multiclass_num,
                    unique=True,
                    seed=self.seed,
                )[0]
            elif self.sample_mode == 'Sample:batch':
                item_indices = tf.reshape(tf.where(tf.reduce_sum(y_true_int, axis=0, keepdims=False) > 0), shape=(-1,))
            else:
                raise MLGBError

            if self.sample_item_distribution_list:
                q = tf.expand_dims(tf.convert_to_tensor(self.sample_item_distribution_list, dtype=tf.float32), axis=0)
            else:
                q = self.get_item_distribution(y_true)

            q = tf.gather(q, indices=item_indices, axis=1)
            y_true = tf.gather(y_true, indices=item_indices, axis=1)
            y_pred = tf.gather(y_pred, indices=item_indices, axis=1)
            y_pred -= tf.math.log(q)

        loss = tf.nn.softmax_cross_entropy_with_logits(
            labels=y_true,
            logits=y_pred,
            axis=-1,
        )
        loss = tf.reduce_mean(loss, axis=None, keepdims=False)
        return loss

    def get_item_distribution(self, y_true, add_bias=1):  # Batch Q like BN, not global Q.
        q_top = tf.reduce_sum(y_true, axis=0, keepdims=True)
        q_bottom = tf.reduce_sum(y_true, axis=None, keepdims=False)
        q = (q_top + add_bias) / (q_bottom + add_bias)  # add 1 avoid log(0) = -inf.
        return q

    def check_shape(self, y_true, y_pred):
        if not (y_true.shape.rank == y_pred.shape.rank == 2):
            raise MLGBError
        if y_true.shape != y_pred.shape:
            raise MLGBError
        if y_true.shape[1] <= 2:
            raise MLGBError
        return

    def get_config(self):  # debug2: get_config is also must. If not, load_model will show bug.
        config = super().get_config()
        config.update(
            {
                'sample_mode': self.sample_mode,
                'sample_num': self.sample_num,
                'sample_item_distribution_list': self.sample_item_distribution_list,
                'sample_fixed_unigram_frequency_list': self.sample_fixed_unigram_frequency_list,
                'sample_fixed_unigram_distortion': self.sample_fixed_unigram_distortion,
                'seed': self.seed,
            }
        )
        return config


class BaseInputsEmbeddingLayer(tf.keras.layers.Layer):
    def __init__(self, user_feature_names,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False, user_inputs_if_embed_dense=True,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Attention', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Attention', pool_seq_axis=2, pool_seq_l2=0.0, pool_seq_initializer=None,
                 user_dnn_hidden_units=(64, 32), dnn_activation=None, dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 tower_embeds_flatten_mode='flatten', tower_embeds_if_l2_norm=True,
                 seed=None):
        super().__init__()
        if tower_embeds_flatten_mode not in ('flatten', 'sum'):
            raise MLGBError

        self.user_dnn_hidden_units = user_dnn_hidden_units
        self.tower_embeds_flatten_mode = tower_embeds_flatten_mode
        self.tower_embeds_if_l2_norm = tower_embeds_if_l2_norm

        self.user_input_fn = InputsLayer(
            feature_names=user_feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=user_inputs_if_multivalued,
            inputs_if_sequential=user_inputs_if_sequential,
            inputs_if_embed_dense=user_inputs_if_embed_dense,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_2d_dim=None,
            embed_cate_if_output2d=False,  #
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        if self.user_dnn_hidden_units:
            self.dnn_fn = DeepNeuralNetworkLayer(
                dnn_hidden_units=user_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed,
            )

    def build(self, input_shape):  # OriginInputs
        if len(input_shape) not in (2, 3, 4):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        _, x = self.user_input_fn(inputs)

        if self.tower_embeds_flatten_mode == 'flatten':
            x = Flatten()(x)
        else:
            x = tf.reduce_sum(x, axis=1, keepdims=False)

        if self.user_dnn_hidden_units:
            x = self.dnn_fn(x)

        if self.tower_embeds_if_l2_norm:
            x = tf.nn.l2_normalize(x, axis=1)
        return x


class CapsuleNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, capsule_num=3, capsule_activation='squash', capsule_l2=0.0, capsule_initializer=None,
                 capsule_interest_num_if_dynamic=False, capsule_input_sequence_pad_mode='pre',
                 capsule_routing_initializer='random_normal', seed=None):
        super().__init__()
        if capsule_input_sequence_pad_mode not in ('pre', 'post'):
            raise MLGBError

        self.capsule_num = capsule_num
        self.capsule_l2 = capsule_l2
        self.capsule_interest_num_if_dynamic = capsule_interest_num_if_dynamic
        self.capsule_input_sequence_pad_mode = capsule_input_sequence_pad_mode
        self.capsule_initializer = InitializerLayer(
            initializer=capsule_initializer,
            activation=capsule_activation,
            seed=seed,
        ).get()
        self.capsule_routing_initializer = InitializerLayer(
            initializer=capsule_routing_initializer,
            activation=None,
            seed=seed,
        ).get()
        self.activation_fn = ActivationLayer(activation=capsule_activation)
        self.mask_fn = MaskLayer(att_if_mask=True)

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        _, seq_len, embed_dim = input_shape
        if self.capsule_interest_num_if_dynamic:
            seq_len = self.get_dynamic_interest_num(seq_len, embed_dim)
            self.seq_len = seq_len

        self.capsule_bilinear_weight = self.add_weight(
            name='capsule_bilinear_weight',
            shape=[seq_len, embed_dim, embed_dim],
            initializer=self.capsule_initializer,
            regularizer=L1L2(0.0, self.capsule_l2),
            trainable=True,
        )
        self.capsule_routing_weight = self.add_weight(
            name='capsule_routing_weight',
            shape=[1, seq_len, embed_dim],
            initializer=self.capsule_routing_initializer,
            regularizer=L1L2(0.0, self.capsule_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        if self.capsule_interest_num_if_dynamic:
            x = x[:, -self.seq_len:, :] if self.capsule_input_sequence_pad_mode == 'pre' else x[:, :self.seq_len, :]

        w = self.capsule_routing_weight
        w_b = self.capsule_bilinear_weight
        for i in range(self.capsule_num):
            w = self.mask_fn(w)
            w = tf.nn.softmax(w, axis=1)
            x_h = tf.einsum('bfe,fee->bfe', x, w_b) * w  # high_level_capsule: x_h = w * w_b * x
            x_h = self.activation_fn(x_h)  # squash

            w_i = tf.einsum('bfe,fee->bfe', x_h, w_b) * x  # routing_logit: w_i = x_h * w_b * x
            w_i = tf.reduce_sum(w_i, axis=0, keepdims=True)
            w = w + w_i  # if `w` isn't be updated(not bp), it's like `w` of RNN.
            x = x_h
        return x

    def get_dynamic_interest_num(self, seq_len, embed_dim):
        seq_k = max(1, min(seq_len, int(numpy.log2(embed_dim))))
        return seq_k












# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy
import pandas
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

from mlgb.utils import (
    get_dense_feature_name_dict,
    get_sparse_feature_name_dict,
    get_padded_sequence,
)
from mlgb.fe import (
    transform_dense_features_to_quantile,
    transform_dense_features_to_discrete,
    transform_sparse_features_to_ordinal,
)
from mlgb.error import MLGBError


pandas.set_option('display.max_columns', None)
pandas.set_option('display.max_rows', None)


def make_sequence_data(
        n_samples=10,
        n_sequence_features=3,
        sequence_if_stable_length=False,
        sequence_max_length=10,
        sequence_choice_if_replace=False,
        vocabulary_size=None,
        ):
    if not (2 < n_sequence_features < 10):
        raise MLGBError

    def sequence_fn(x):
        sequence_min_length = max(1, sequence_max_length - n_sequence_features)
        sequence_size = numpy.random.randint(sequence_min_length, sequence_max_length + 1)
        x = numpy.random.choice(x, size=sequence_size, replace=sequence_choice_if_replace)
        return x

    def id_fn(x, i):  # (cate_id > shop_id > good_id)
        x = numpy.array(x)
        x = numpy.round(x // ((vocabulary_size ** (1 / n_sequence_features)) ** i))
        x = numpy.where(x >= 1, x, 1)
        x = x.astype(numpy.int32)
        return x

    vocabulary_size = vocabulary_size if vocabulary_size else int(1e3)
    vocabulary_size = vocabulary_size if vocabulary_size >= sequence_max_length else sequence_max_length

    sequence_id_pool = list(range(1, vocabulary_size))  # omit building dictionary
    df = pandas.DataFrame({'sequence': [sequence_id_pool] * n_samples})
    df['sequence'] = df['sequence'].apply(sequence_fn)
    for i in range(n_sequence_features):
        df[f'sequence_{i}'] = df['sequence'].apply(lambda x: id_fn(x, i))

    df = df.iloc[:, 1:]
    if sequence_if_stable_length:
        df = get_padded_sequence(df, mode='pandas', sequence_max_length=sequence_max_length)
    return df  # (b, f, s): (b, s, f) -> (b, f, s) -> (b, f, s, e) -> seq:(b, s, e) or mv:(b, f, e).


def make_fake_data(
        n_samples=1000,
        task='binary',
        n_classes=2,
        class_weights=(0.9,),
        n_numerical_features=30,
        n_categorical_features=5,
        n_sequential_features=5,
        sequence_if_stable_length=False,
        sequence_max_length=5,
        sequence_choice_if_replace=False,
        multiclass_features_if_add_item_id=False,
        multitask_cvr=0.1,
        vocabulary_size=None,
        shuffle=True,
        seed=None
        ):

    if isinstance(task, str) and task not in ('binary', 'regression'):
        raise MLGBError
    if isinstance(task, (tuple, list)) and tuple(task) != ('binary', 'binary',):
        raise MLGBError

    def regression_y_fn(y):
        betas = numpy.random.uniform(low=0.0, high=0.5, size=y.shape)
        y = numpy.where(y >= 0.5, y-betas, y+betas)
        return y

    def ctr_cvr_binary_y_fn(y, cvr=0.5):
        y_ctr = y.reshape(-1, 1)
        betas = numpy.random.choice([0, 1], p=(1 - cvr, cvr), size=y_ctr.shape, replace=True)
        y_cvr = numpy.where((y_ctr == 1) & (betas == 1), 1, 0)
        y_ctr_cvr = numpy.concatenate([y_ctr, y_cvr], axis=1)
        return y_ctr_cvr

    vocabulary_size = vocabulary_size if vocabulary_size else int(n_samples // 10)
    category_pool = [f'id_{i}' for i in range(1, vocabulary_size+1)]

    if n_classes <= 2:
        n_features = n_numerical_features
    else:
        n_features = max([n_numerical_features, n_classes + 3])

    numerical_features, y = make_classification(
        n_classes=n_classes,
        weights=class_weights,
        n_samples=n_samples,
        n_features=n_features,  # >= n_classes + 3 if n_classes > 2 else `any`.
        n_informative=n_classes,
        n_redundant=2,
        n_repeated=0,
        shuffle=shuffle,
        random_state=seed,
    )

    if n_features > n_numerical_features:
        numerical_features = numerical_features[:, :n_numerical_features]

    if task == 'regression':
        y = regression_y_fn(y)
    if isinstance(task, (tuple, list)) and tuple(task) == ('binary', 'binary',):
        y = ctr_cvr_binary_y_fn(y, multitask_cvr)

    numerical_features_names = [f'numerical:c{i + 1}' for i in range(n_numerical_features)]
    categorical_features_names = [f'categorical:c{i + 1}' for i in range(n_categorical_features)]
    sequential_features_names = [f'sequential:c{i + 1}' for i in range(n_sequential_features)]

    numerical_features = pandas.DataFrame(numerical_features, columns=numerical_features_names)
    categorical_features = numpy.random.choice(category_pool, size=[n_samples, n_categorical_features], replace=True)
    categorical_features = pandas.DataFrame(categorical_features, columns=categorical_features_names)
    if multiclass_features_if_add_item_id and n_classes > 2:
        categorical_features['y'] = y

    sequential_features = make_sequence_data(
        n_samples=n_samples,
        n_sequence_features=n_sequential_features,
        sequence_if_stable_length=sequence_if_stable_length,
        sequence_max_length=sequence_max_length,
        sequence_choice_if_replace=sequence_choice_if_replace,
        vocabulary_size=vocabulary_size,
    )
    sequential_features.columns = sequential_features_names

    data = {
        'x': pandas.concat([numerical_features, categorical_features, sequential_features], axis=1),
        'y': y,
        'is_multiclass': True if n_classes > 2 else False,
        'numerical_features_names': numerical_features_names,
        'categorical_features_names': categorical_features_names,
        'sequential_features_names': sequential_features_names,
        'sequence_max_length': sequence_max_length,
        # 'sequence_vocabulary_size': vocabulary_size,
    }
    return data


def make_input_data(data, task='binary', test_size=0.15, inputs_if_2_groups=False, multiclass_item_features_if_only_item_id=False,  seed=None):
    sparse_ord_features_from_numerical = transform_dense_features_to_discrete(
        numerical_features=data['x'][data['numerical_features_names']],
        n_bins=20,
        strategy='quantile',
        encode='ordinal',
        seed=seed
    )
    sparse_ord_features_from_categorical = transform_sparse_features_to_ordinal(
        categorical_features=data['x'][data['categorical_features_names']]
    )

    dense_features = transform_dense_features_to_quantile(
        numerical_features=data['x'][data['numerical_features_names']],
        n_quantiles=100,
        seed=seed
    )
    sparse_features = pandas.concat([sparse_ord_features_from_categorical, sparse_ord_features_from_numerical], axis=1)
    sequential_features = data['x'][data['sequential_features_names']]

    if tuple(task) == ('binary', 'binary',):
        label = data['y']
        x_data = pandas.concat([dense_features, sparse_features, sequential_features], axis=1)
        x_cols = list(x_data.columns)
        y_cols = ['y_ctr', 'y_cvr']
        x_data[y_cols] = label
        x_y_data = x_data
        train_data, test_data = train_test_split(x_y_data, test_size=test_size, random_state=seed)
        train_data, y_train = train_data[x_cols], train_data[y_cols].to_numpy()
        test_data, y_test = test_data[x_cols], test_data[y_cols].to_numpy()
    else:
        label = data['y']
        x_data = pandas.concat([dense_features, sparse_features, sequential_features], axis=1)
        train_data, test_data, y_train, y_test = train_test_split(x_data, label, stratify=label, test_size=test_size, random_state=seed)

    if inputs_if_2_groups:
        if multiclass_item_features_if_only_item_id and data['is_multiclass']:
            y_train_data = pandas.DataFrame({'y': y_train}).astype(numpy.int32)
            y_test_data = pandas.DataFrame({'y': y_test}).astype(numpy.int32)

            x1_train = tuple([
                train_data[dense_features.columns].to_numpy().astype(numpy.float32),
                train_data[sparse_features.columns].to_numpy().astype(numpy.int32),
                numpy.array(train_data[sequential_features.columns].to_numpy().tolist()).astype(numpy.int32),
            ])
            x1_test = tuple([
                test_data[dense_features.columns].to_numpy().astype(numpy.float32),
                test_data[sparse_features.columns].to_numpy().astype(numpy.int32),
                numpy.array(test_data[sequential_features.columns].to_numpy().tolist()).astype(numpy.int32),
            ])

            x2_train = tuple([
                y_train_data['y'].to_numpy().reshape(-1, 1).astype(numpy.float32),
                y_train_data['y'].to_numpy().reshape(-1, 1).astype(numpy.int32),
            ])
            x2_test = tuple([
                y_test_data['y'].to_numpy().reshape(-1, 1).astype(numpy.float32),
                y_test_data['y'].to_numpy().reshape(-1, 1).astype(numpy.int32),
            ])
            x_train = (x1_train, x2_train,)
            x_test = (x1_test, x2_test,)

            feature_1_names = tuple([
                tuple([
                    get_dense_feature_name_dict(name) for name in dense_features.columns
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=x_data[name].nunique(),
                        input_length=1,
                    ) for name in sparse_features.columns
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=len(set(it for item in x_data[name].to_numpy() for it in item)),
                        input_length=data['sequence_max_length'],
                    ) for name in sequential_features.columns
                ]),
            ])
            feature_2_names = tuple([
                tuple([
                    get_dense_feature_name_dict(name) for name in y_train_data.columns
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=y_train_data[name].nunique(),
                        input_length=1,
                    ) for name in y_train_data.columns
                ]),
            ])
            feature_names = (feature_1_names, feature_2_names,)
        else:
            id_dense_half = len(dense_features.columns) // 2
            id_sparse_half = len(sparse_features.columns) // 2
            id_sequential_half = len(sequential_features.columns) // 2

            x1_train = tuple([
                train_data[dense_features.columns[:id_dense_half]].to_numpy().astype(numpy.float32),
                train_data[sparse_features.columns[:id_sparse_half]].to_numpy().astype(numpy.int32),
                numpy.array(train_data[sequential_features.columns[:id_sequential_half]].to_numpy().tolist()).astype(numpy.int32),
            ])
            x1_test = tuple([
                test_data[dense_features.columns[:id_dense_half]].to_numpy().astype(numpy.float32),
                test_data[sparse_features.columns[:id_sparse_half]].to_numpy().astype(numpy.int32),
                numpy.array(test_data[sequential_features.columns[:id_sequential_half]].to_numpy().tolist()).astype(numpy.int32),
            ])

            x2_train = tuple([
                train_data[dense_features.columns[id_dense_half:]].to_numpy().astype(numpy.float32),
                train_data[sparse_features.columns[id_sparse_half:]].to_numpy().astype(numpy.int32),
                numpy.array(train_data[sequential_features.columns[id_sequential_half:]].to_numpy().tolist()).astype(numpy.int32),
            ])
            x2_test = tuple([
                test_data[dense_features.columns[id_dense_half:]].to_numpy().astype(numpy.float32),
                test_data[sparse_features.columns[id_sparse_half:]].to_numpy().astype(numpy.int32),
                numpy.array(test_data[sequential_features.columns[id_sequential_half:]].to_numpy().tolist()).astype(numpy.int32),
            ])
            x_train = (x1_train, x2_train,)
            x_test = (x1_test, x2_test,)

            feature_1_names = tuple([
                tuple([
                    get_dense_feature_name_dict(name) for name in dense_features.columns[:id_dense_half]
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=x_data[name].nunique(),
                        input_length=1,
                    ) for name in sparse_features.columns[:id_sparse_half]
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=len(set(it for item in x_data[name].to_numpy() for it in item)),
                        input_length=data['sequence_max_length'],
                    ) for name in sequential_features.columns[:id_sequential_half]
                ]),
            ])
            feature_2_names = tuple([
                tuple([
                    get_dense_feature_name_dict(name) for name in dense_features.columns[id_dense_half:]
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=x_data[name].nunique(),
                        input_length=1,
                    ) for name in sparse_features.columns[id_sparse_half:]
                ]),
                tuple([
                    get_sparse_feature_name_dict(
                        feature_name=name,
                        feature_nunique=len(set(it for item in x_data[name].to_numpy() for it in item)),
                        input_length=data['sequence_max_length'],
                    ) for name in sequential_features.columns[id_sequential_half:]
                ]),
            ])
            feature_names = (feature_1_names, feature_2_names,)
    else:
        x_train = tuple([
            train_data[dense_features.columns].to_numpy().astype(numpy.float32),
            train_data[sparse_features.columns].to_numpy().astype(numpy.int32),
            numpy.array(train_data[sequential_features.columns].to_numpy().tolist()).astype(numpy.int32),
        ])
        x_test = tuple([
            test_data[dense_features.columns].to_numpy().astype(numpy.float32),
            test_data[sparse_features.columns].to_numpy().astype(numpy.int32),
            numpy.array(test_data[sequential_features.columns].to_numpy().tolist()).astype(numpy.int32),
        ])
        feature_names = tuple([
            tuple([
                get_dense_feature_name_dict(name) for name in dense_features.columns
            ]),
            tuple([
                get_sparse_feature_name_dict(
                    feature_name=name,
                    feature_nunique=x_data[name].nunique(),
                    input_length=1,
                ) for name in sparse_features.columns
            ]),
            tuple([
                get_sparse_feature_name_dict(
                    feature_name=name,
                    feature_nunique=len(set(it for item in x_data[name].to_numpy() for it in item)),
                    input_length=data['sequence_max_length'],
                ) for name in sequential_features.columns
            ]),
        ])

    if task == 'binary' or tuple(task) == ('binary', 'binary',):
        y_train = y_train.astype(numpy.int32)
        y_test = y_test.astype(numpy.int32)
    else:
        y_train = y_train.astype(numpy.float32)
        y_test = y_test.astype(numpy.float32)
    return feature_names, (x_train, y_train), (x_test, y_test)


def get_label_data(n_samples=1000, task='binary', n_classes=2, class_weights=(0.9,), test_size=0.15, seed=None,
                   multitask_cvr=0.1, inputs_if_2_groups=False,
                   multiclass_item_features_if_only_item_id=False, multiclass_features_if_add_item_id=False,):
    if isinstance(task, str) and task not in ('binary', 'regression'):
        raise MLGBError
    if isinstance(task, (tuple, list)) and tuple(task) != ('binary', 'binary',):
        raise MLGBError
    if multiclass_features_if_add_item_id and multiclass_item_features_if_only_item_id and n_classes > 2:
        multiclass_features_if_add_item_id = False

    data = make_fake_data(
        n_samples=n_samples,
        task=task,
        n_classes=n_classes,
        class_weights=class_weights,
        n_numerical_features=50,
        n_categorical_features=10,
        n_sequential_features=8,
        sequence_if_stable_length=True,
        sequence_max_length=10,
        sequence_choice_if_replace=False,
        multiclass_features_if_add_item_id=multiclass_features_if_add_item_id,
        multitask_cvr=multitask_cvr,
        vocabulary_size=None,
        shuffle=True,
        seed=seed,
    )
    out = make_input_data(
        data=data,
        task=task,
        inputs_if_2_groups=inputs_if_2_groups,
        multiclass_item_features_if_only_item_id=multiclass_item_features_if_only_item_id,
        test_size=test_size,
        seed=seed,
    )
    return out


def get_regression_label_data(n_samples=1000, longtail_weight=0.9, test_size=0.15, inputs_if_2_groups=False, seed=None):
    data = get_label_data(
        n_samples=n_samples,
        inputs_if_2_groups=inputs_if_2_groups,
        task='regression',
        n_classes=2,
        class_weights=(longtail_weight,),
        test_size=test_size,
        seed=seed,
    )
    return data


def get_binary_label_data(n_samples=1000, negative_class_weight=0.9, test_size=0.15, inputs_if_2_groups=False, seed=None):
    data = get_label_data(
        n_samples=n_samples,
        inputs_if_2_groups=inputs_if_2_groups,
        task='binary',
        n_classes=2,
        class_weights=(negative_class_weight,),
        test_size=test_size,
        seed=seed,
    )
    return data


def get_multitask_label_data(n_samples=1000, negative_class_weight=0.9, multitask_cvr=0.5, test_size=0.15, seed=None):
    data = get_label_data(
        n_samples=n_samples,
        task=('binary', 'binary',),
        inputs_if_2_groups=False,
        multitask_cvr=multitask_cvr,
        n_classes=2,
        class_weights=(negative_class_weight,),
        test_size=test_size,
        seed=seed,
    )
    return data


def get_multiclass_label_data(n_samples=1000, n_classes=100, test_size=0.15, inputs_if_2_groups=False,
                              multiclass_item_features_if_only_item_id=False, seed=None):
    data = get_label_data(
        n_samples=n_samples,
        task='binary',
        n_classes=n_classes,
        class_weights=None,
        test_size=test_size,
        inputs_if_2_groups=inputs_if_2_groups,
        multiclass_features_if_add_item_id=False,
        multiclass_item_features_if_only_item_id=multiclass_item_features_if_only_item_id and inputs_if_2_groups and n_classes > 2,
        seed=seed,
    )
    return data


if __name__ == '__main__':
    # data = make_fake_data(n_samples=20)
    # binary_label_data = get_binary_label_data(50)
    seq_data = make_sequence_data(
        n_samples=10,
        n_sequence_features=3,
        sequence_if_stable_length=True,
        sequence_max_length=5,
        sequence_choice_if_replace=False,
    )
    # b_data = get_binary_label_data(1000)
    # r_data = get_regression_label_data(1000)
    # m_data = get_multiclass_label_data(1000, 10, inputs_if_2_groups=True, multiclass_item_features_if_only_item_id=True)

    # y_train = b_data[1][1]
    # y_test = b_data[2][1]
    # print(sum(y_train) / len(y_train))
    # print(sum(y_test) / len(y_test))

    # ctr_data = get_multitask_label_data(1000)
    # y_train = ctr_data[1][1]
    # y_test = ctr_data[2][1]
    # print(sum(y_train[:, 0]) / len(y_train))
    # print(sum(y_test[:, 0]) / len(y_test))
    # print(sum(y_train[:, 1]) / len(y_train))
    # print(sum(y_test[:, 1]) / len(y_test))




# coding=utf-8
# author=uliontse

import numpy
import pandas
import tensorflow as tf
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix, classification_report, roc_auc_score, ConfusionMatrixDisplay

from mlgb import get_model
from mlgb.tf.components.retrieval import SampledSoftmaxLossLayer
from mlgb.data import get_multiclass_label_data
from mlgb.utils import check_filepath


def get_item_distribution_list(y_true, add_bias=1):
    q_top = tf.reduce_sum(y_true, axis=0, keepdims=True)
    q_bottom = tf.reduce_sum(y_true, axis=None, keepdims=False)
    q = (q_top + add_bias) / (q_bottom + add_bias)  # add 1 avoid log(0) = -inf.
    return q.numpy().reshape(-1).tolist()


def plot_confusion_matrix(y_confusion_matrix, save_dir='.', dpi=800):
    n_classes = y_confusion_matrix.shape[0]
    plt.rcParams.update({'font.size': round(100 / n_classes + 1)})
    _, ax = plt.subplots(figsize=(8.0, 6.0), dpi=dpi)
    display = ConfusionMatrixDisplay(y_confusion_matrix)
    display.plot(ax=ax)
    plt.savefig(f'{save_dir}/plot_confusion_matrix.png', dpi=dpi)
    plt.show()
    return


if __name__ == '__main__':
    model_name = 'YoutubeDNN'
    sample_mode = 'Sample:batch'
    print(f'model_name: {model_name}')

    tmp_dir = '.tmp'
    model_dir = f'{tmp_dir}/{model_name}_tf'
    log_dir = f'{model_dir}/log_dir'
    save_model_dir = f'{model_dir}/save_model'
    check_filepath(tmp_dir, model_dir, log_dir, save_model_dir)

    seed = 0
    n_classes = 100
    feature_names, (x_train, y_train), (x_test, y_test) = get_multiclass_label_data(
        n_samples=int(2e4),
        n_classes=n_classes,
        inputs_if_2_groups=True,
        multiclass_item_features_if_only_item_id=True,
        seed=seed,
    )
    y_int_train, y_int_test = y_train, y_test
    y_train = tf.keras.utils.to_categorical(y_train, num_classes=n_classes)
    y_test = tf.keras.utils.to_categorical(y_test, num_classes=n_classes)
    print([len(names) for names in feature_names[0]])
    print([len(names) for names in feature_names[1]])

    print('get_item_distribution.')
    global_y_true = numpy.concatenate([y_train, y_test], axis=0)
    global_q_list = get_item_distribution_list(global_y_true)

    model = get_model(
        feature_names=feature_names,
        model_name=model_name,
        task=f'multiclass:{n_classes}',
        aim='matching',
        lang='tf',
        seed=seed,

        model_result_temperature_ratio=None,
        sample_mode=sample_mode,
        sample_num=None,
        sample_item_distribution_list=global_q_list,
        sample_fixed_unigram_frequency_list=global_q_list,
        user_dnn_hidden_units=(256, 128),
    )
    model.compile(
        loss=model.sampled_softmax_loss,
        optimizer=tf.optimizers.Nadam(learning_rate=1e-3),
        metrics=[tf.metrics.AUC(multi_label=True, from_logits=True if sample_mode else False)],
    )
    history = model.fit(
        x=x_train,
        y=y_train,
        batch_size=64,
        epochs=100,
        validation_split=0.2,
        callbacks=[
            tf.keras.callbacks.TensorBoard(log_dir=log_dir),
            tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=10, min_delta=1e-3),
        ]
    )

    print(model.summary())
    print(pandas.DataFrame(history.history))

    test_evaluate = model.evaluate(x=x_test, y=y_test, batch_size=64, return_dict=True)
    print(test_evaluate)

    y_pred_prob = model.predict(x_test)
    y_pred_prob = tf.nn.softmax(y_pred_prob, axis=-1).numpy()
    auc = roc_auc_score(
        y_true=y_test,
        y_score=y_pred_prob,
        average='macro',
        multi_class='ovr',
    )
    print('auc:', auc)

    y_int_pred = numpy.argmax(y_pred_prob, axis=-1).reshape(-1)
    print(y_int_pred.shape)
    print(y_int_test.shape)

    print(cm := confusion_matrix(y_int_test, y_int_pred))
    plot_confusion_matrix(cm, save_dir=model_dir)
    print(classification_report(y_int_test, y_int_pred))

    # ranking model to deploy:
    model.save(filepath=save_model_dir)
    local_model = tf.keras.models.load_model(
        filepath=save_model_dir,
        custom_objects={'SampledSoftmaxLossLayer': SampledSoftmaxLossLayer}
    )

    y_local_pred = local_model.predict(x_test)
    y_model_pred = model.predict(x_test)
    print('y_local_pred == y_model_pred:', numpy.allclose(y_local_pred, y_model_pred))
    print(numpy.abs(y_local_pred - y_model_pred).max())

    # matching model to deploy:
    user_inputs, item_inputs = x_test

    if model_name == 'NCF':
        user_mf_embeddings = local_model.user_mf_embedding_fn(user_inputs)
        item_mf_embeddings = local_model.item_mf_embedding_fn(item_inputs)
        user_dnn_embeddings = local_model.user_dnn_embedding_fn(user_inputs)
        item_dnn_embeddings = local_model.item_dnn_embedding_fn(item_inputs)

        y = local_model.ncf_fn([user_mf_embeddings, item_mf_embeddings, user_dnn_embeddings, item_dnn_embeddings])
    else:
        user_embeddings = local_model.user_embedding_fn(user_inputs)
        item_embeddings = local_model.item_embedding_fn(item_inputs)

        y = user_embeddings * item_embeddings

    y_match_pred = local_model.task_fn(y)
    y_match_pred = y_match_pred.numpy()
    print('y_match_pred == y_model_pred:', numpy.allclose(y_match_pred, y_model_pred))
    print(numpy.abs(y_match_pred - y_model_pred).max())




# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    tf,
    Flatten,
)
from mlgb.tf.inputs import InputsLayer
from mlgb.tf.components.linears import (
    DeepNeuralNetworkLayer,
)
from mlgb.tf.components.retrieval import (
    BaseInputsEmbeddingLayer,
    CapsuleNetworkLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'DeepStructuredSemanticModelUserEmbeddingLayer',
    'DeepStructuredSemanticModelItemEmbeddingLayer',
    'NeuralCollaborativeFilteringLayer',
    'MultiInterestNetworkWithDynamicRoutingLayer',
    'UserInputs_MultiInterestNetworkWithDynamicRoutingLayer',
]


class DeepStructuredSemanticModelUserEmbeddingLayer(BaseInputsEmbeddingLayer):
    def __init__(self, user_feature_names,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False, user_inputs_if_embed_dense=True,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Attention', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Attention', pool_seq_axis=2, pool_seq_l2=0.0, pool_seq_initializer=None,
                 user_dnn_hidden_units=(64, 32), dnn_activation='tanh', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 tower_embeds_flatten_mode='flatten', tower_embeds_if_l2_norm=True,
                 seed=None):
        super().__init__(
            user_feature_names,
            user_inputs_if_multivalued, user_inputs_if_sequential, user_inputs_if_embed_dense,
            embed_dim, embed_l2, embed_initializer,
            pool_mv_mode, pool_mv_axis, pool_mv_l2, pool_mv_initializer,
            pool_seq_mode, pool_seq_axis, pool_seq_l2, pool_seq_initializer,
            user_dnn_hidden_units, dnn_activation, dnn_dropout,
            dnn_if_bn, dnn_if_ln, dnn_l2, dnn_initializer,
            tower_embeds_flatten_mode, tower_embeds_if_l2_norm,
            seed,
        )


class DeepStructuredSemanticModelItemEmbeddingLayer(BaseInputsEmbeddingLayer):
    def __init__(self, item_feature_names,
                 item_inputs_if_multivalued=False, item_inputs_if_sequential=False, item_inputs_if_embed_dense=True,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Attention', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Attention', pool_seq_axis=2, pool_seq_l2=0.0, pool_seq_initializer=None,
                 item_dnn_hidden_units=(64, 32), dnn_activation='tanh', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 tower_embeds_flatten_mode='flatten', tower_embeds_if_l2_norm=True,
                 seed=None):
        super().__init__(
            item_feature_names,
            item_inputs_if_multivalued, item_inputs_if_sequential, item_inputs_if_embed_dense,
            embed_dim, embed_l2, embed_initializer,
            pool_mv_mode, pool_mv_axis, pool_mv_l2, pool_mv_initializer,
            pool_seq_mode, pool_seq_axis, pool_seq_l2, pool_seq_initializer,
            item_dnn_hidden_units, dnn_activation, dnn_dropout,
            dnn_if_bn, dnn_if_ln, dnn_l2, dnn_initializer,
            tower_embeds_flatten_mode, tower_embeds_if_l2_norm,
            seed,
        )


class NeuralCollaborativeFilteringLayer(tf.keras.layers.Layer):
    def __init__(self, dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 4:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == input_shape[2].rank == input_shape[3].rank == 2):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_user_mf, x_item_mf, x_user_dnn, x_item_dnn = inputs

        x_dnn = tf.concat([x_user_dnn, x_item_dnn], axis=1)
        dnn_outputs = self.dnn_fn(x_dnn)
        mf_outputs = x_user_mf * x_item_mf

        ncf_outputs = tf.concat([mf_outputs, dnn_outputs], axis=1)
        return ncf_outputs


class MultiInterestNetworkWithDynamicRoutingLayer(tf.keras.layers.Layer):
    def __init__(self, capsule_num=3, capsule_activation='squash', capsule_l2=0.0, capsule_initializer=None,
                 capsule_interest_num_if_dynamic=False, capsule_input_sequence_pad_mode='pre',
                 capsule_routing_initializer='random_normal',
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.caps_net_fn = CapsuleNetworkLayer(
            capsule_num=capsule_num,
            capsule_activation=capsule_activation,
            capsule_l2=capsule_l2,
            capsule_initializer=capsule_initializer,
            capsule_interest_num_if_dynamic=capsule_interest_num_if_dynamic,
            capsule_input_sequence_pad_mode=capsule_input_sequence_pad_mode,
            capsule_routing_initializer=capsule_routing_initializer,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn_list = [Flatten() for _ in range(2)]

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_user_fea, x_user_seq = inputs

        x_user_seq = self.caps_net_fn(x_user_seq)
        x_user_seq = self.flatten_fn_list[0](x_user_seq)
        x_user_fea = self.flatten_fn_list[1](x_user_fea)
        x = tf.concat([x_user_fea, x_user_seq], axis=1)
        x = self.dnn_fn(x)
        return x


class UserInputs_MultiInterestNetworkWithDynamicRoutingLayer(tf.keras.layers.Layer):
    def __init__(self, user_feature_names,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=True,  #
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=2, pool_seq_l2=0.0, pool_seq_initializer=None,
                 tower_embeds_if_l2_norm=True,
                 capsule_num=3, capsule_activation='squash', capsule_l2=0.0, capsule_initializer=None,
                 capsule_interest_num_if_dynamic=False, capsule_input_sequence_pad_mode='pre',
                 capsule_routing_initializer='random_normal',
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.tower_embeds_if_l2_norm = tower_embeds_if_l2_norm

        self.user_input_fn = InputsLayer(
            feature_names=user_feature_names,
            inputs_mode='Inputs:sequential',
            inputs_if_multivalued=user_inputs_if_multivalued,
            inputs_if_sequential=user_inputs_if_sequential,
            inputs_if_embed_dense=True,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_2d_dim=None,
            embed_cate_if_output2d=False,  #
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,  #
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        self.user_mind_fn = MultiInterestNetworkWithDynamicRoutingLayer(
            capsule_num=capsule_num,
            capsule_activation=capsule_activation,
            capsule_l2=capsule_l2,
            capsule_initializer=capsule_initializer,
            capsule_interest_num_if_dynamic=capsule_interest_num_if_dynamic,
            capsule_input_sequence_pad_mode=capsule_input_sequence_pad_mode,
            capsule_routing_initializer=capsule_routing_initializer,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
    
    def build(self, input_shape):  # OriginInputs
        if len(input_shape) not in (2, 3, 4):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        _, user_fea_3d_inputs, user_seq_3d_inputs = self.user_input_fn(inputs)
        
        x = self.user_mind_fn([user_fea_3d_inputs, user_seq_3d_inputs])
        if self.tower_embeds_if_l2_norm:
            x = tf.nn.l2_normalize(x, axis=1)
        return x











































# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy
import tensorflow as tf


Embedding = tf.keras.layers.Embedding
Dense = tf.keras.layers.Dense
Activation = tf.keras.layers.Activation
L1L2 = tf.keras.regularizers.L1L2
Dropout = tf.keras.layers.Dropout
BatchNormalization = tf.keras.layers.BatchNormalization
LayerNormalization = tf.keras.layers.LayerNormalization
Flatten = tf.keras.layers.Flatten
Conv1D = tf.keras.layers.Conv1D
Conv2D = tf.keras.layers.Conv2D
MaxPool1D = tf.keras.layers.MaxPool1D
MaxPool2D = tf.keras.layers.MaxPool2D
GRU = tf.keras.layers.GRU


InputsModeList = ['Inputs:feature', 'Inputs:sequential']
PoolModeList = ['Pooling:max', 'Pooling:average', 'Pooling:sum']
MVPoolModeList = ['Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
BiGRUModeList = ['Frontward', 'Backward', 'Frontward+Backward', 'Frontward-Backward', 'Frontward*Backward', 'Frontward,Backward']
SeqRecPointwiseModeList = ['Add', 'LabelAttention', 'Add&LabelAttention']

EDCNModeList = ['EDCN:pointwise_addition', 'EDCN:hadamard_product', 'EDCN:concatenation', 'EDCN:attention_pooling']
PNNModeList = ['PNN:inner_product', 'PNN:outer_product', 'PNN:both']
BLModeList = ['Bilinear:field_all', 'Bilinear:field_each', 'Bilinear:field_interaction']
FBIModeList = [
    'FM', 'FM3D',
    'FFM', 'HOFM', 'FwFM', 'FEFM', 'FvFM', 'FmFM', 'AFM',
    'PNN:inner_product', 'PNN:outer_product', 'PNN:both',
    'Bilinear:field_all', 'Bilinear:field_each', 'Bilinear:field_interaction',
]

SampleModeList = [
    'Sample:all', 'Sample:batch',
    'Sample:uniform', 'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram',
]



# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.torch.configs import (
    random,
    numpy,
    torch,
)
from mlgb.error import MLGBError


__all__ = [
    'SeedLayer',
    'IdentityLayer',
    'FlattenLayer',
    'FlattenAxesLayer',
    'TransposeLayer',
    'RegularizationLayer',
    'BatchNormalizationLayer',
    'LayerNormalizationLayer',
    'ActivationLayer',
    'InitializerLayer',
    'OneHotLayer',
    'SparseEmbeddingLayer',
    'DenseEmbeddingLayer',
    'PositionalEncoding',
    'BiasEncoding',
    'MaskLayer',
    'SimplePoolingLayer',
    'MultiValuedPoolingLayer',
    'KMaxPoolingLayer',
]


class SeedLayer:
    def __init__(self, seed=None):
        self.seed = seed

    def reset(self):
        random.seed(self.seed)
        numpy.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        return


class RegularizationLayer:
    def __init__(self, model, l1=0.0, l2=0.0):
        if not (l1 >= 0 and l2 >= 0):
            raise MLGBError

        self.model = model
        self.l1 = l1
        self.l2 = l2

    def get_l1l2_loss(self):
        x = torch.concat([w.detach().view(-1) for w in self.model.parameters()], dim=0)  # no_grad

        l1l2_loss = 0.0
        if self.l1 > 0:
            l1l2_loss += self.p_norm(x, p=1, if_root=False)
        if self.l2 > 0:
            l1l2_loss += self.p_norm(x, p=2, if_root=False)
        return l1l2_loss

    def p_norm(self, x, p=2, if_root=True):  # torch.norm()
        if not p >= 1:
            raise MLGBError

        lp = torch.sum(torch.abs(x) ** p, dim=None, keepdim=False) ** ((1 / p) if if_root else 1)
        return lp


class IdentityLayer(torch.nn.Module):
    def __init__(self, if_stop_gradient=False, device='cpu'):
        super().__init__()
        self.if_stop_gradient = if_stop_gradient
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x
        x = x.detach() if self.if_stop_gradient else x  # torch.tensor(x, requires_grad=False)
        return x


class TransposeLayer(torch.nn.Module):
    def __init__(self, perm=(1, 2), device='cpu'):
        super().__init__()
        if len(perm) != 2:
            raise MLGBError

        self.perm = perm
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.transpose(x, *self.perm)
        return x


class FlattenLayer(torch.nn.Module):
    def __init__(self, device='cpu'):
        super().__init__()
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.reshape(x, shape=[x.shape[0], -1])
        return x


class FlattenAxesLayer(torch.nn.Module):
    def __init__(self, axes=(1, 2), device='cpu'):
        super().__init__()
        if len(axes) != 2:
            raise MLGBError
        if 0 in axes:
            raise MLGBError

        self.axes = axes
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 4:
                raise MLGBError

            _, self.p1, self.p2, self.p3 = x.shape
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if list(self.axes) == [1, 2]:
            x = torch.reshape(x, shape=[-1, self.p1 * self.p2, self.p3])
        elif list(self.axes) == [2, 3]:
            x = torch.reshape(x, shape=[-1, self.p1, self.p2 * self.p3])
        elif list(self.axes) == [1, 3]:
            x = torch.reshape(x, shape=[-1, self.p1 * self.p3, self.p2])
        else:
            raise MLGBError
        return x


class BatchNormalizationLayer(torch.nn.Module):
    def __init__(self, axis=1, eps=1e-3, momentum=0.99, if_affine=True, device='cpu'):
        super().__init__()
        # tf_moving_average_decay = 1 - tf_momentum = torch_momentum = torch_moving_average_factor.
        self.eps = eps
        self.momentum = 1.0 - momentum  # use tf_momentum.
        self.if_affine = if_affine
        self.dims = list(range(3))
        self.dim = self.dims[axis]
        self.perm = [self.dim, 1]  #
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.transpose(x, *self.perm)
        x = torch.nn.BatchNorm1d(
            num_features=x.shape[1],
            eps=self.eps,
            momentum=self.momentum,
            affine=self.if_affine,
            device=self.device,
        )(x)
        x = torch.transpose(x, *self.perm)
        return x


class LayerNormalizationLayer(torch.nn.Module):
    def __init__(self, axis=1, eps=1e-3, if_affine=True, device='cpu'):
        super().__init__()
        self.eps = eps
        self.if_affine = if_affine
        self.dims = list(range(3))
        self.dim = self.dims[axis]
        self.perm = [self.dim, -1]  #
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.transpose(x, *self.perm)
        x = torch.nn.LayerNorm(
            normalized_shape=[x.shape[-1]],
            eps=self.eps,
            elementwise_affine=self.if_affine,
            bias=self.if_affine,
            device=self.device,
        )(x)
        x = torch.transpose(x, *self.perm)
        return x


class DiceLayer(torch.nn.Module):
    def __init__(self, dice_if_official_bn=True, device='cpu'):
        super().__init__()
        self.epsilon = 1e-8
        self.alpha = torch.nn.init.ones_(torch.nn.parameter.Parameter(
            data=torch.empty(
                size=[],  # scalar is freedom.
                # device=self.device,
            ),
            requires_grad=True,
        ))

        self.official_bn_fn = BatchNormalizationLayer(
            axis=1,
            eps=self.epsilon,
            if_affine=False,
            device=device,
        )
        self.bn_fn = self.official_bn_fn if dice_if_official_bn else self.unofficial_bn_fn
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        # swish(x) = silu(x) = sigmoid(x) * x
        # prelu(x) = tf.where(x > 0, x, self.alpha * x)
        # dice(x) = tf.where(x > 0, sigmoid(x) * x, (1 - sigmoid(x)) * self.alpha * x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        p = self.bn_fn(x)
        p = torch.sigmoid(p)
        x = p * x + (1 - p) * self.alpha * x
        return x

    def unofficial_bn_fn(self, x):
        diff = x - torch.mean(x, dim=0, keepdim=True)
        var = torch.var(x, dim=0, keepdim=True)
        x = diff / torch.sqrt(var + self.epsilon)
        return x


class SquashLayer(torch.nn.Module):
    def __init__(self, device='cpu'):
        super().__init__()
        self.alpha = 1.0
        self.beta = 1e-8
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x
        l1 = self.p_norm(x, p=1, if_root=True)
        l2 = self.p_norm(x, p=2, if_root=True)
        x = x * (l2 / (l2 + self.alpha)) * (1 / (l1 + self.beta))
        return x

    def p_norm(self, x, p=2, if_root=True):  # torch.norm()
        lp = torch.sum(torch.abs(x) ** p, dim=None, keepdim=False) ** ((1 / p) if if_root else 1)
        return lp


class ActivationLayer(torch.nn.Module):
    def __init__(self, activation=None, device='cpu'):
        super().__init__()
        self.activation = activation
        self.activation_map = {
            'relu': torch.nn.ReLU(),
            'gelu': torch.nn.GELU(),
            'silu': torch.nn.SiLU(),
            'prelu': torch.nn.PReLU(),
            'selu': torch.nn.SELU(),
            'tanh': torch.nn.Tanh(),
            'sigmoid': torch.nn.Sigmoid(),
            'dice': DiceLayer(dice_if_official_bn=True, device=device),
            'squash': SquashLayer(device=device),
        }
        if self.activation is None:
            self.activation_fn = IdentityLayer(device=device)
        elif isinstance(self.activation, torch.nn.Module):
            self.activation_fn = self.activation
        else:
            self.activation_fn = self.activation_map[self.activation]
        
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x
        x = self.activation_fn(x)
        return x


class InitializerLayer:  # Not Layer
    def __init__(self, initializer=None, activation=None):
        self.initializer = initializer
        self.activation = activation
        self.initializer_map = {
            'glorot_normal': torch.nn.init.xavier_normal_,
            'glorot_uniform': torch.nn.init.xavier_uniform_,
            'xavier_normal': torch.nn.init.xavier_normal_,
            'xavier_uniform': torch.nn.init.xavier_uniform_,
            'he_normal': torch.nn.init.kaiming_normal_,  # he_scale=2, lecun_scale=1.
            'he_uniform': torch.nn.init.kaiming_uniform_,
            'kaiming_normal': torch.nn.init.kaiming_normal_,
            'kaiming_uniform': torch.nn.init.kaiming_uniform_,
            'random_normal': torch.nn.init.normal_,
            'random_uniform': torch.nn.init.uniform_,
            'truncated_normal': torch.nn.init.trunc_normal_,
            'orthogonal': torch.nn.init.orthogonal_,
            'zeros': torch.nn.init.zeros_,
            'ones': torch.nn.init.ones_,
        }
        self.activation_initializer_map = {
            'relu': self.initializer_map['he_normal'],
            'gelu': self.initializer_map['he_normal'],
            'silu': self.initializer_map['he_normal'],
            'prelu': self.initializer_map['he_normal'],
            'selu': self.initializer_map['he_normal'],  # lecun_normal
            'tanh': self.initializer_map['glorot_normal'],
            'sigmoid': self.initializer_map['glorot_normal'],
            'dice': self.initializer_map['glorot_normal'],
            'squash': self.initializer_map['random_normal'],
        }

    def get(self):
        if callable(self.initializer):  # types.FunctionType
            initializer_fn = self.initializer
        elif not (self.initializer in self.initializer_map or self.activation in self.activation_initializer_map):
            initializer_fn = self.initializer_map['glorot_normal']
        elif self.initializer:
            initializer_fn = self.initializer_map[self.initializer]
        else:
            initializer_fn = self.activation_initializer_map[self.activation]
        return initializer_fn


class OneHotLayer(torch.nn.Module):
    def __init__(self, sparse_feature_names, onehot_dim=None, device='cpu'):
        super().__init__()
        self.sparse_feature_names = sparse_feature_names
        self.onehot_dim = onehot_dim
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 2:
                raise MLGBError

            self.fields_width = x.shape[1]
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x
        x = x.to(dtype=torch.int64) if x.dtype != torch.int64 else x  # Must LongTensor.

        sparse_tensor = x
        onehot_tensor_list = []
        for i in range(self.fields_width):
            depth = self.onehot_dim if self.onehot_dim else self.sparse_feature_names[i]['feature_nunique']
            sparse_i_tensor = sparse_tensor[:, i]
            onehot_i_tensor = torch.nn.functional.one_hot(sparse_i_tensor, num_classes=depth)
            onehot_tensor_list.append(onehot_i_tensor)
        onehot_tensor = torch.concat(onehot_tensor_list, dim=1)
        return onehot_tensor


class SparseEmbeddingLayer(torch.nn.Module):
    def __init__(self, sparse_feature_names, embed_dim=None, embed_if_output2d=False, device='cpu'):
        super().__init__()
        self.sparse_feature_names = sparse_feature_names
        self.embed_dim = embed_dim
        self.embed_if_output2d = embed_if_output2d
        self.embed_fn_map = {
            f'{i}': torch.nn.Embedding(
                num_embeddings=feat_dict['embed_feature_nunique'],
                embedding_dim=feat_dict['embed_dim'] if not self.embed_dim else self.embed_dim,
                padding_idx=0,
                device=device,
            )
            for i, feat_dict in enumerate(self.sparse_feature_names)
        }
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim not in (2, 3):
                raise MLGBError

            self.fields_width = x.shape[1]
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x
        x = x.to(dtype=torch.int32) if x.dtype not in (torch.int32, torch.int64) else x  # Must LongTensor or IntTensor.

        sparse_tensor = x
        embed_tensor_list = [self.embed_fn_map[f'{i}'](sparse_tensor[:, i, ...]) for i in range(self.fields_width)]
        d23_tensor = torch.concat(embed_tensor_list, dim=1)

        if self.embed_if_output2d:
            return d23_tensor, d23_tensor

        if not self.embed_dim:
            raise MLGBError

        d34_tensor = torch.stack(embed_tensor_list, dim=1)
        return d23_tensor, d34_tensor


class DenseEmbeddingLayer(torch.nn.Module):
    def __init__(self, embed_dim=32, embed_initializer=None, device='cpu'):
        super().__init__()
        if not embed_dim:
            raise MLGBError

        self.embed_dim = embed_dim
        self.initializer_fn = InitializerLayer(
            initializer=embed_initializer,
            activation=None,
        ).get()
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 2:
                raise MLGBError

            self.embed_weight = self.initializer_fn(torch.nn.parameter.Parameter(
                data=torch.empty(
                    size=[x.shape[1], self.embed_dim],
                    device=self.device,
                ),
                requires_grad=True,
            ))
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.einsum('bf,fe->bfe', x, self.embed_weight)
        return x


class PositionalEncoding(torch.nn.Module):
    def __init__(self, att_if_pe=True, device='cpu'):
        super().__init__()
        self.att_if_pe = att_if_pe
        self.wave_length = 1e4  # 1e4 * pi
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            _, f, e = x.shape

            pos_f = numpy.arange(f).reshape(-1, 1)
            pos_e = numpy.arange(e) // 2 * 2
            pos_x = pos_f / numpy.power(self.wave_length, pos_e / e)  # (f, e)

            pos_x[:, 0::2] = numpy.sin(pos_x[:, 0::2])
            pos_x[:, 1::2] = numpy.cos(pos_x[:, 1::2])

            self.pe = torch.as_tensor(pos_x.reshape(1, f, e), dtype=torch.float32, device=self.device)
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if self.att_if_pe:
            x += self.pe
        return x


class BiasEncoding(torch.nn.Module):
    def __init__(self, if_bias=True, bias_initializer='zeros', device='cpu'):
        super().__init__()
        self.if_bias = if_bias
        self.initializer_fn = InitializerLayer(
            initializer=bias_initializer,
            activation=None,
        ).get()
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim < 2:
                raise MLGBError

            if self.if_bias:
                self.bias = self.initializer_fn(torch.nn.parameter.Parameter(
                    data=torch.empty(
                        size=x.shape[1:],
                        device=self.device,
                    ),
                    requires_grad=True,
                ))

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if self.if_bias:
            x += self.bias
        return x


class MaskLayer(torch.nn.Module):
    def __init__(self, att_if_mask=True, device='cpu'):
        super().__init__()
        self.min_inf = -4294967295  # float32: -2 ** 32 + 1
        self.att_if_mask = att_if_mask
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def forward(self, x):
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if self.att_if_mask:
            x = torch.where(torch.not_equal(x, 0), x, self.min_inf)
        return x
        

class SimplePoolingLayer(torch.nn.Module):
    def __init__(self, pool_mode='Pooling:average', pool_axis=-1, pool_axis_if_keep=False, device='cpu'):
        super().__init__()
        if pool_mode not in ('Pooling:max', 'Pooling:average', 'Pooling:sum'):
            raise MLGBError

        self.pool_mode = pool_mode
        self.pool_axis = pool_axis
        self.pool_axis_if_keep = pool_axis_if_keep
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim not in (2, 3, 4):
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = self.pool_fn(x, self.pool_mode, self.pool_axis, self.pool_axis_if_keep)
        return x

    def pool_fn(self, x, pool_mode, pool_axis, pool_axis_if_keep):
        pool_fn_map = {
            'Pooling:average': lambda x: torch.mean(x, dim=pool_axis, keepdim=pool_axis_if_keep),
            'Pooling:sum': lambda x: torch.sum(x, dim=pool_axis, keepdim=pool_axis_if_keep),
            'Pooling:max': lambda x: torch.max(x, dim=pool_axis, keepdim=pool_axis_if_keep),
        }
        x = pool_fn_map[pool_mode](x)
        return x


class MultiValuedPoolingLayer(torch.nn.Module):
    def __init__(self, pool_mode='Attention', pool_axis=2, pool_if_output2d=False, pool_initializer=None, device='cpu'):
        super().__init__()
        if pool_mode not in ('Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum'):
            raise MLGBError
        if pool_axis not in (1, 2, 3, -1, -2, -3):
            raise MLGBError

        self.pool_mode = pool_mode
        self.pool_axis = pool_axis
        self.pool_if_output2d = pool_if_output2d
        self.initializer_fn = InitializerLayer(
            initializer=pool_initializer,
            activation=None,
        ).get()
        if self.pool_mode.startswith('Pooling:'):
            self.pool_fn = SimplePoolingLayer(
                pool_mode=pool_mode,
                pool_axis=pool_axis,
                device=device,
            )
        self.flatten_fn = FlattenLayer(device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 4:
                raise MLGBError

            _, self.fields_width, self.sequence_length, _ = x.shape
            if self.pool_mode in ('Weighted', 'Attention'):
                self.pool_weight = self.initializer_fn(torch.nn.parameter.Parameter(
                    data=torch.empty(
                        size=[self.fields_width, self.sequence_length],
                        device=self.device,
                    ),
                    requires_grad=True,
                ))
            self.built = True
        return

    def forward(self, x):
        self.build(x)  # (batch_size, fields_width, sequence_length, embed_dim)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if self.pool_mode.startswith('Pooling:'):
            x = self.pool_fn(x)
        elif self.pool_mode in ('Weighted', 'Attention'):
            x_w = torch.einsum('ijkl,jk->ijkl', x, self.pool_weight)
            if self.pool_mode == 'Attention':
                x_w = MaskLayer(att_if_mask=True)(x_w)
                x_w = x * torch.softmax(x_w, dim=self.pool_axis)
            x = torch.sum(x_w, dim=self.pool_axis, keepdim=False)
        else:
            raise MLGBError

        if self.pool_if_output2d:
            x = self.flatten_fn(x)
        return x  # (b, f, e) or (b, f*e)


class KMaxPoolingLayer(torch.nn.Module):
    def __init__(self, pool_axis=-1, device='cpu'):
        super().__init__()
        if pool_axis == 0:
            raise MLGBError

        self.axis = pool_axis
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim not in (3, 4):
                raise MLGBError

            self.built = True
        return

    def forward(self, x, k):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.topk(x, k=k, dim=self.axis, largest=True, sorted=True)[0]
        return x









